#!/usr/bin/env python3
# NES NTD compressor,
# Copyright (C) 2015  Johnathan Roatch
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import collections
import functools
import io

"""
Stream format:
    0nnnnnnn        : 128-N Run of BG.
    100nnnnn ...    : 32-N literal bytes.
    10111111 xx     : set BG to X. (initially BG = 0)
    10111110 xx yy  : Unused. (decodes as 2 bytes)
    101nnnnn xx yy  : For 32-N times, write alternately X and Y.
    11011111 xx     : Unused. (decodes as 1 byte)
    110nnnnn xx     : 32-N incrementing run starting at X.
    11111111        : End stream.
    111nnnnn xx     : 32-N run of X.
"""

def decompress_file(in_file, out_file, bs=8192, ignore_unused_codes=False):
    in_file_iter = itertools.chain.from_iterable(
        iter(functools.partial(in_file.read, bs), b'')
    )
    bg_byte = 0x00
    for cmd in in_file_iter:
        if cmd == 0b11111111:
            break
        elif cmd == 0b10111111:
            bg_byte = next(in_file_iter)
        elif cmd & 0b10000000 == 0:
            count = 128 - (cmd & 0b01111111)
            out_file.write( bytes(bg_byte for i in range(count)) )
        elif ignore_unused_codes and cmd == 0b10111110:
            pass
        elif ignore_unused_codes and cmd == 0b11011111:
            pass
        else:
            count = 32 - (cmd & 0b00011111)
            cmd_type = (cmd & 0b01100000) >> 5
            run_bytes = bytearray(count)
            if cmd_type == 0:
                run_bytes = bytes([next(in_file_iter) for i in range(count)])
            elif cmd_type == 1:
                x = next(in_file_iter)
                y = next(in_file_iter)
                for i in range(count):
                    run_bytes[i] = x
                    x, y = y, x
            elif cmd_type == 2:
                x = next(in_file_iter)
                for i in range(count):
                    run_bytes[i] = x
                    x = (x+1)&0xff
            elif cmd_type == 3:
                x = next(in_file_iter)
                for i in range(count):
                    run_bytes[i] = x
            out_file.write(run_bytes)
    else:
        return False
    return True

# TODO: Rewrite this to also compress the cases of interleaved common runs.
def compress_file(in_file, out_file, bs=8192, write_end_marker=True):
    input_buffer = collections.deque(maxlen=4)
    literals = bytearray()
    run_length = 0
    run_value = 0
    run_type = None
    bg_byte = 0x00

    in_file_iter = itertools.chain.from_iterable(
        iter(functools.partial(in_file.read, bs), b'')
    )

    for byte in itertools.chain(in_file_iter, [None, None]):
        if run_length > 0:
            if run_type == 'Run' and (byte != run_value or run_length >= 128):
                if run_value == bg_byte:
                    out_file.write(bytes(
                        [(128 - run_length) | 0b00000000]
                    ))
                elif run_value != bg_byte and run_length > 32:
                    bg_byte = run_value
                    out_file.write(bytes(
                        [0b10111111, bg_byte, (128 - run_length) | 0b00000000]
                    ))
                else:
                    out_file.write(bytes(
                        [(32 - run_length) | 0b11100000, run_value]
                    ))
                if byte is not None:
                    input_buffer.append(byte)
                run_length = 0
                run_type = None
            elif ((run_type == 'Inc') and (run_length == 2) and
                    (byte == (run_value)&0xff)):
                run_type = 'Dbl'
                run_value = [(run_value+1)&0xff, run_value]
                run_length = 3
            elif ((run_type == 'Inc') and
                    ((byte != (run_value+run_length)&0xff) or run_length >= 32)):
                out_file.write(bytes(
                    [(32 - run_length) | 0b11000000, run_value]
                ))
                if byte is not None:
                    input_buffer.append(byte)
                run_length = 0
                run_type = None
            elif run_type == 'Dbl' and (byte != run_value[0] or run_length >= 32):
                if run_length & 0x01:
                    run_value[0], run_value[1] = run_value[1], run_value[0]
                out_file.write(bytes(
                    [(32 - run_length) | 0b10100000, run_value[0], run_value[1]]
                ))
                if byte is not None:
                    input_buffer.append(byte)
                run_length = 0
                run_type = None
            else:
                run_length += 1
                if run_type == 'Dbl':
                    run_value[0], run_value[1] = run_value[1], run_value[0]
        else:
            if byte is not None:
                input_buffer.append(byte)
            if (len(input_buffer) == 2) and (input_buffer[0] == input_buffer[1]):
                run_type = 'Run'
            elif (len(input_buffer) == 2) and ((input_buffer[0]+1)&0xff == input_buffer[1]):
                run_type = 'Inc'
            elif (len(input_buffer) == 3) and (input_buffer[0] == input_buffer[2]):
                run_type = 'Dbl'
            elif (len(input_buffer) >= 4) and (input_buffer[0] == input_buffer[2]) and (input_buffer[1] == input_buffer[3]):
                run_type = 'Dbl'
            elif len(input_buffer) >= 3:
                if input_buffer[-3] == input_buffer[-2] == input_buffer[-1]:
                    run_type = 'Run'
                elif (input_buffer[-3]+2)&0xff == (input_buffer[-2]+1)&0xff == input_buffer[-1]:
                    run_type = 'Inc'
            if run_type == 'Run':
                run_length = 1
                while ((len(input_buffer) >= 2) and
                        (input_buffer[-2] == input_buffer[-1])):
                    run_length += 1
                    input_buffer.pop()
                run_value = input_buffer.pop()
            elif run_type == 'Inc':
                run_length = 1
                while ((len(input_buffer) >= 2) and
                        ((input_buffer[-2]+1)&0xff == input_buffer[-1])):
                    run_length += 1
                    input_buffer.pop()
                run_value = input_buffer.pop()
            elif run_type == 'Dbl':
                run_length = 2
                run_value = [input_buffer[-2], input_buffer[-1]]
                while ((len(input_buffer) >= 3) and
                        (input_buffer[-3] == input_buffer[-1])):
                    run_length += 1
                    input_buffer.pop()
                input_buffer.pop()
                input_buffer.pop()
            if len(input_buffer) >= 4:
                literals.append(input_buffer.popleft())
            if run_length > 0 or len(literals) >= 32 or byte is None:
                while (len(input_buffer) > 0 and
                        len(literals) < 32):
                    literals.append(input_buffer.popleft())
                if len(literals) > 0:
                    if all(i == bg_byte for i in literals):
                        out_file.write(bytes(
                            [(128 - len(literals)) | 0b00000000]
                        ))
                    else:
                        out_file.write(bytes(
                            [(32 - len(literals)) | 0b10000000]
                        ))
                        out_file.write(literals)
                    literals.clear()
    if write_end_marker:
        out_file.write(b'\xff')

def compress(data, write_end_marker=True):
    with io.BytesIO(data) as input_file:
        with io.BytesIO() as output_file:
            compress_file(input_file, output_file, write_end_marker=write_end_marker)
            return output_file.getvalue()

def decompress(data, ignore_unused_codes=False):
    with io.BytesIO(data) as input_file:
        with io.BytesIO() as output_file:
            decompress_file(input_file, output_file, ignore_unused_codes=ignore_unused_codes)
            return output_file.getvalue()


if __name__=='__main__':
    import sys
    import argparse

    parser = argparse.ArgumentParser(description='NES NTD compressor', usage='%(prog)s [-d] [options] -o output [-i] input')
    parser.add_argument('--version', action='version', version='%(prog)s 2.0')
    parser.add_argument('input', metavar='files', help='Input files', nargs='*')
    parser.add_argument('-o', '--output', metavar='file', help='Output file')
    parser.add_argument('-f', '--force',  help='Force overwrite of output file.', action='store_true')
    parser.add_argument('-d', '--decompress', help='Decode the input file.', action='store_true')
    parser.add_argument('--omit-end-marker', help='Don\'t Write the 0xff byte at the end of the compressed output.', action='store_true')
    options = parser.parse_args()

    if '-' not in options.input and not sys.stdin.isatty():
        options.input.append('-')

    if options.output is None and not sys.stdout.isatty():
        options.output = '-'

    if not options.input or not options.output:
        sys.exit(parser.format_help())

    if options.output == '-':
        output_file = sys.stdout.buffer
    elif options.force:
        output_file = open(options.output, 'wb')
    else:
        output_file = open(options.output, 'xb')

    for fn in options.input:
        if fn == '-':
            input_file = sys.stdin.buffer
        else:
            input_file = open(fn, 'rb')

        if options.decompress:
            decompress_file(input_file, output_file)
            # TODO:
            # Deal with decoding errors, indicated by StopIteration exceptions.
            # Also indicate if the sream ended with EOF or the end marker.
        else:
            # Multiple input files are concatenated.
            # So we'll save the end marker untill all the files are procsessed.
            compress_file(input_file, output_file, write_end_marker=False)

        if fn != '-':
            input_file.close()

    if not options.decompress and not options.omit_end_marker:
        output_file.write(b'\xff')

    if options.output != '-':
        output_file.close()

